@@ -1,26 +1,20 @@
-#=======================================================================
+# Code in the PDF::API2::Basic::PDF namespace was originally copied from the
+# Text::PDF distribution.
 #
-#   THIS IS A REUSED PERL MODULE, FOR PROPER LICENCING TERMS SEE BELOW:
+# Copyright Martin Hosken <Martin_Hosken@sil.org>
 #
-#   Copyright Martin Hosken <Martin_Hosken@sil.org>
-#
-#   Modified for PDF::API2 by Alfred Reibenschuh <alfredreibenschuh@gmx.net>
-#
-#   No warranty or expression of effectiveness, least of all regarding
-#   anyone's safety, is implied in this software or documentation.
-#
-#   This specific module is licensed under the Perl Artistic License.
-#
-#=======================================================================
+# Martin Hosken's code may be used under the terms of the MIT license.
+# Subsequent versions of the code have the same license as PDF::API2.
+
 package PDF::API2::Basic::PDF::File;
 
 use strict;
 
-our $VERSION = '2.030'; # VERSION
+our $VERSION = '2.044'; # VERSION
 
 =head1 NAME
 
-PDF::API2::Basic::PDF::File - Holds the trailers and cross-reference tables for a PDF file
+PDF::API2::Basic::PDF::File - Low-level PDF file access
 
 =head1 SYNOPSIS
 
@@ -140,7 +134,7 @@ is in PDF which contains the location of the previous cross-reference table.
 
 =cut
 
-use Scalar::Util qw(blessed);
+use Scalar::Util qw(blessed weaken);
 
 use vars qw($cr $irreg_char $reg_char $ws_char $delim_char %types);
 
@@ -150,6 +144,9 @@ $reg_char = '[^][<>{}()/% \t\r\n\f\0]';
 $irreg_char = '[][<>{}()/% \t\r\n\f\0]';
 $cr = '\s*(?:\015|\012|(?:\015\012))';
 
+my $re_comment = qr/(?:\%[^\r\n]*)/;
+my $re_whitespace = qr/(?:[ \t\r\n\f\0]|$re_comment)/;
+
 %types = (
     'Page'  => 'PDF::API2::Basic::PDF::Page',
     'Pages' => 'PDF::API2::Basic::PDF::Pages',
@@ -157,6 +154,7 @@ $cr = '\s*(?:\015|\012|(?:\015\012))';
 
 my $readDebug = 0;
 
+use Carp;
 use IO::File;
 
 # Now for the basic PDF types
@@ -225,8 +223,13 @@ sub open {
         $fh = $filename;
     }
     else {
-        die "File '$filename' does not exist !" unless -f $filename;
-        $fh = IO::File->new(($update ? '+' : '') . "<$filename") || return;
+        die "File '$filename' does not exist"  unless -f $filename;
+        die "File '$filename' is not readable" unless -r $filename;
+        if ($update) {
+            die "File '$filename' is not writable" unless -w $filename;
+        }
+        $fh = IO::File->new(($update ? '+' : '') . "<$filename")
+            || die "Error opening '$filename': $!";
         $self->{' INFILE'} = $fh;
         if ($update) {
             $self->{' update'} = 1;
@@ -237,8 +240,8 @@ sub open {
     binmode $fh, ':raw';
     $fh->seek(0, 0);            # go to start of file
     $fh->read($buffer, 255);
-    unless ($buffer =~ m/^\%PDF\-1\.(\d)+\s*$cr/mo) {
-        die "$filename not a PDF file version 1.x";
+    unless ($buffer =~ /^\%PDF\-([12]\.\d+)\s*$cr/m) {
+        croak "$filename does not appear to be a valid PDF";
     }
     $self->{' version'} = $1;
 
@@ -263,6 +266,88 @@ sub open {
     return $self;
 }
 
+=head2 $p->version($version)
+
+Gets/sets the PDF version (e.g. 1.4)
+
+=cut
+
+sub version {
+    my $self = shift();
+
+    if (@_) {
+        my $version = shift();
+        croak "Invalid version $version" unless $version =~ /^([12]\.[0-9]+)$/;
+        $self->header_version($version);
+        if ($version >= 1.4) {
+            $self->trailer_version($version);
+        }
+        else {
+            delete $self->{'Root'}->{'Version'};
+            $self->out_obj($self->{'Root'});
+        }
+        return $version;
+    }
+
+    my $header_version = $self->header_version();
+    my $trailer_version = $self->trailer_version();
+    return $trailer_version if $trailer_version > $header_version;
+    return $header_version;
+}
+
+=head2 $version = $p->header_version($version)
+
+Gets/sets the PDF version stored in the file header.
+
+=cut
+
+sub header_version {
+    my $self = shift();
+
+    if (@_) {
+        my $version = shift();
+        croak "Invalid version $version" unless $version =~ /^([12]\.[0-9]+)$/;
+        $self->{' version'} = $version;
+    }
+
+    return $self->{' version'};
+}
+
+=head2 $version = $p->trailer_version($version)
+
+Gets/sets the PDF version stored in the document catalog.
+
+=cut
+
+sub trailer_version {
+    my $self = shift();
+
+    if (@_) {
+        my $version = shift();
+        croak "Invalid version $version" unless $version =~ /^([12]\.[0-9]+)$/;
+        $self->{'Root'}->{'Version'} = PDFName($version);
+        $self->out_obj($self->{'Root'});
+        return $version;
+    }
+
+    return unless $self->{'Root'}->{'Version'};
+    $self->{'Root'}->{'Version'}->realise();
+    return $self->{'Root'}->{'Version'}->val();
+}
+
+=head2 $prev_version = $p->require_version($version)
+
+Ensures that the PDF version is at least C<$version>.
+
+=cut
+
+sub require_version {
+    my ($self, $min_version) = @_;
+    my $current_version = $self->version();
+    $self->version($min_version) if $current_version < $min_version;
+    return $current_version;
+}
+
 =head2 $p->release()
 
 Releases ALL of the memory used by the PDF document and all of its
@@ -295,6 +380,11 @@ sub release {
         delete $self->{$key};
     }
 
+    # PDFs with highly-interconnected page trees or outlines can hit Perl's
+    # recursion limit pretty easily, so disable the warning for this specific
+    # loop.
+    no warnings 'recursion';
+
     while (my $item = shift @tofree) {
         if (blessed($item) and $item->can('release')) {
             $item->release(1);
@@ -327,13 +417,6 @@ sub append_file {
 
     my $fh = $self->{' INFILE'};
 
-    # hack to upgrade pdf-version number to support
-    # requested features in higher versions than
-    # the pdf was originally created.
-    my $version = $self->{' version'} || 4;
-    $fh->seek(0, 0);
-    $fh->print("%PDF-1.$version\n");
-
     my $tdict = PDFDict();
     $tdict->{'Prev'} = PDFNum($self->{' loc'});
     $tdict->{'Info'} = $self->{'Info'};
@@ -396,12 +479,45 @@ sub create_file {
     }
 
     $self->{' OUTFILE'} = $fh;
-    $fh->print('%PDF-1.' . ($self->{' version'} || '2') . "\n");
+    $fh->print('%PDF-' . ($self->{' version'} // '1.4') . "\n");
     $fh->print("%\xC6\xCD\xCD\xB5\n");   # and some binary stuff in a comment
     return $self;
 }
 
 
+=head2 $p->clone_file($fname)
+
+Creates a copy of the input file at the specified filename and sets it as the
+output file for future writes.  A file handle may be passed instead of a
+filename.
+
+=cut
+
+sub clone_file {
+    my ($self, $filename) = @_;
+    my $fh;
+
+    $self->{' fname'} = $filename;
+    if (ref $filename) {
+        $fh = $filename;
+    }
+    else {
+        $fh = IO::File->new(">$filename") || die "Unable to open $filename for writing";
+        binmode($fh,':raw');
+    }
+
+    $self->{' OUTFILE'} = $fh;
+
+    my $in = $self->{' INFILE'};
+    $in->seek(0, 0);
+    my $data;
+    while (not $in->eof()) {
+        $in->read($data, 1024 * 1024);
+        $fh->print($data);
+    }
+    return $self;
+}
+
 =head2 $p->close_file
 
 Closes up the open file for output by outputting the trailer etc.
@@ -421,7 +537,7 @@ sub close_file {
     $tdict->{'Size'} = $self->{'Size'} || PDFNum(1);
     $tdict->{'Prev'} = PDFNum($self->{' loc'}) if $self->{' loc'};
     if ($self->{' update'}) {
-        foreach my $key (grep ($_ !~ m/^[\s\-]/, keys %$self)) {
+        foreach my $key (grep { $_ !~ m/^[\s\-]/ } keys %$self) {
             $tdict->{$key} = $self->{$key} unless defined $tdict->{$key};
         }
 
@@ -473,17 +589,23 @@ sub readval {
             if ($str =~ s|^/($reg_char+)||) {
                 my $key = PDF::API2::Basic::PDF::Name::name_to_string($1, $self);
                 ($value, $str) = $self->readval($str, %opts);
-                $result->{$key} = $value;
+                unless ((ref($value) // '') eq 'PDF::API2::Basic::PDF::Null') {
+                    $result->{$key} = $value;
+                }
             }
             elsif ($str =~ s|^/$ws_char+||) {
                 # fixes a broken key problem of acrobat. -- fredo
                 ($value, $str) = $self->readval($str, %opts);
-                $result->{'null'} = $value;
+                unless ((ref($value) // '') eq 'PDF::API2::Basic::PDF::Null') {
+                    $result->{'null'} = $value;
+                }
             }
             elsif ($str =~ s|^//|/|) {
                 # fixes again a broken key problem of illustrator/enfocus. -- fredo
                 ($value, $str) = $self->readval($str, %opts);
-                $result->{'null'} = $value;
+                unless ((ref($value) // '') eq 'PDF::API2::Basic::PDF::Null') {
+                    $result->{'null'} = $value;
+                }
             }
             else {
                 die "Invalid dictionary key";
@@ -524,10 +646,10 @@ sub readval {
     }
 
     # Indirect Object
-    elsif ($str =~ m/^([0-9]+)$ws_char+([0-9]+)$ws_char+R/s) {
+    elsif ($str =~ m/^([0-9]+)(?:$ws_char|$re_comment)+([0-9]+)(?:$ws_char|$re_comment)+R/s) {
         my $num = $1;
         $value = $2;
-        $str =~ s/^([0-9]+)$ws_char+([0-9]+)$ws_char+R//s;
+        $str =~ s/^([0-9]+)(?:$ws_char|$re_comment)+([0-9]+)(?:$ws_char|$re_comment)+R//s;
         unless ($result = $self->test_obj($num, $value)) {
             $result = PDF::API2::Basic::PDF::Objind->new();
             $result->{' objnum'} = $num;
@@ -535,17 +657,21 @@ sub readval {
             $self->add_obj($result, $num, $value);
         }
         $result->{' parent'} = $self;
-        $result->{' realised'} = 0;
+        weaken $result->{' parent'};
+
+        # Removed to address changes being lost when an indirect object is realised twice
+        # $result->{' realised'} = 0;
+
         # gdj: FIXME: if any of the ws chars were crs, then the whole
         # string might not have been read.
     }
 
     # Object
-    elsif ($str =~ m/^([0-9]+)$ws_char+([0-9]+)$ws_char+obj/s) {
+    elsif ($str =~ m/^([0-9]+)(?:$ws_char|$re_comment)+([0-9]+)(?:$ws_char|$re_comment)+obj/s) {
         my $obj;
         my $num = $1;
         $value = $2;
-        $str =~ s/^([0-9]+)$ws_char+([0-9]+)$ws_char+obj//s;
+        $str =~ s/^([0-9]+)(?:$ws_char|$re_comment)+([0-9]+)(?:$ws_char|$re_comment)+obj//s;
         ($obj, $str) = $self->readval($str, %opts);
         if ($result = $self->test_obj($num, $value)) {
             $result->merge($obj);
@@ -645,21 +771,33 @@ sub readval {
     }
 
     # Boolean
-    elsif ($str =~ m/^(true|false)$irreg_char/) {
+    elsif ($str =~ m/^(true|false)($irreg_char|$)/) {
         $value = $1;
         $str =~ s/^(?:true|false)//;
         $result = PDF::API2::Basic::PDF::Bool->from_pdf($value);
     }
 
     # Number
-    elsif ($str =~ m/^([+-.0-9]+)$irreg_char/) {
+    elsif ($str =~ m/^([+-.0-9]+)($irreg_char|$)/) {
         $value = $1;
         $str =~ s/^([+-.0-9]+)//;
+
+        # If $str only consists of whitespace (or is empty), call update to
+        # see if this is the beginning of an indirect object or reference
+        if ($update and ($str =~ /^$re_whitespace*$/s or $str =~ /^$re_whitespace+[0-9]+$re_whitespace*$/s)) {
+            $str =~ s/^$re_whitespace+/ /s;
+            $str =~ s/$re_whitespace+$/ /s;
+            $str = update($fh, $str);
+            if ($str =~ m/^$re_whitespace*([0-9]+)$re_whitespace+(?:R|obj)/s) {
+                return $self->readval("$value $str", %opts);
+            }
+        }
+
         $result = PDF::API2::Basic::PDF::Number->from_pdf($value);
     }
 
     # Null
-    elsif ($str =~ m/^null$irreg_char/) {
+    elsif ($str =~ m/^null($irreg_char|$)/) {
         $str =~ s/^null//;
         $result = PDF::API2::Basic::PDF::Null->new;
     }
@@ -697,35 +835,68 @@ Returns a fully read object of given number and generation in this file
 
 sub read_objnum {
     my ($self, $num, $gen, %opts) = @_;
+    croak 'Undefined object number in call to read_objnum($num, $gen)' unless defined $num;
+    croak 'Undefined object generation in call to read_objnum($num, $gen)' unless defined $gen;
+    croak "Invalid object number '$num' in call to read_objnum" unless $num =~ /^[0-9]+$/;
+    croak "Invalid object generation '$gen' in call to read_objnum" unless $gen =~ /^[0-9]+$/;
 
     my $object_location = $self->locate_obj($num, $gen) || return;
     my $object;
 
-    if (ref $object_location)
-    {
-        # Compressed object.
-        my $src = $self->read_objnum($object_location->[0], 0, %opts);
-        die 'Cannot find the compressed object stream' unless $src;
+    # Compressed object
+    if (ref($object_location)) {
+        my ($object_stream_num, $object_stream_pos) = @{$object_location};
 
-        $src->read_stream(1) if $src->{' nofilt'};
+        my $object_stream = $self->read_objnum($object_stream_num, 0, %opts);
+        die 'Cannot find the compressed object stream' unless $object_stream;
 
-        my $map = substr($src->{' stream'}, 0, $src->{'First'}->val);
-        my $objects = substr($src->{' stream'}, $src->{'First'}->val);
-        my @mappings = split(/\s+/, $map);
-        my $count = scalar(@mappings);
+        $object_stream->read_stream() if $object_stream->{' nofilt'};
 
-        my $index = $object_location->[1] * 2;
+        # An object stream starts with pairs of integers containing object numbers and
+        # stream offsets relative to the First key
+        my $fh;
+        my $pairs;
+        unless ($object_stream->{' streamfile'}) {
+            $pairs = substr($object_stream->{' stream'}, 0, $object_stream->{'First'}->val);
+        }
+        else {
+            CORE::open($fh, '<', $object_stream->{' streamfile'});
+            read($fh, $pairs, $object_stream->{'First'}->val());
+        }
+        my @map = split /\s+/, $pairs;
 
-        if ($mappings[$index] != $num)
-        {
-            die "Objind $num does not exist at index $index";
+        # Find the offset of the object in the stream
+        my $index = $object_stream_pos * 2;
+        die "Objind $num does not exist at index $index" unless $map[$index] == $num;
+        my $start = $map[$index + 1];
+
+        # Unless this is the last object in the stream, its length is determined by the
+        # offset of the next object
+        my $last_object_in_stream = $map[-2];
+        my $length;
+        if ($last_object_in_stream == $num) {
+            if ($object_stream->{' stream'}) {
+                $length = length($object_stream->{' stream'}) - $object_stream->{'First'}->val() - $start;
+            }
+            else {
+                $length = (-s $object_stream->{' streamfile'}) - $object_stream->{'First'}->val() - $start;
+            }
+        }
+        else {
+            my $next_start = $map[$index + 3];
+            $length = $next_start - $start;
         }
 
-        my $start = $mappings[++$index];
-        $index += 2;
-
-        my $length = $index > $count ? length($objects) : $mappings[$index];
-        my $stream = "$num 0 obj" . substr($objects, $start, $length);
+        # Read the object from the stream
+        my $stream = "$num 0 obj ";
+        unless ($object_stream->{' streamfile'}) {
+            $stream .= substr($object_stream->{' stream'}, $object_stream->{'First'}->val() + $start, $length);
+        }
+        else {
+            seek($fh, $object_stream->{'First'}->val() + $start, 0);
+            read($fh, $stream, $length, length($stream));
+            close $fh;
+        }
 
         ($object) = $self->readval($stream, %opts, update => 0);
         return $object;
@@ -778,7 +949,6 @@ sub new_obj {
                 }
                 else {
                     $res = $self->test_obj($i, $ng) || $self->add_obj(PDF::API2::Basic::PDF::Objind->new(), $i, $ng);
-                    $tdict->{' xref'}{$i}[0] = $tdict->{' xref'}{$i}[0];
                     $self->out_obj($res);
                     return $res;
                 }
@@ -818,7 +988,8 @@ sub out_obj {
     # in the hash) (which is super-fast).
     unless (exists $self->{' outlist_cache'}{$obj}) {
         push @{$self->{' outlist'}}, $obj;
-        $self->{' outlist_cache'}{$obj}++;
+        # weaken $self->{' outlist'}->[-1];
+        $self->{' outlist_cache'}{$obj} = 1;
     }
     return $obj;
 }
@@ -852,8 +1023,8 @@ sub remove_obj {
     delete $self->{' objects'}{$objind->uid()};
     delete $self->{' outlist_cache'}{$objind};
     delete $self->{' printed_cache'}{$objind};
-    @{$self->{' outlist'}} = grep($_ ne $objind, @{$self->{' outlist'}});
-    @{$self->{' printed'}} = grep($_ ne $objind, @{$self->{' printed'}});
+    @{$self->{' outlist'}} = grep { $_ ne $objind } @{$self->{' outlist'}};
+    @{$self->{' printed'}} = grep { $_ ne $objind } @{$self->{' printed'}};
     $self->{' objcache'}{$objind->{' objnum'}, $objind->{' objgen'}} = undef
         if $self->{' objcache'}{$objind->{' objnum'}, $objind->{' objgen'}} eq $objind;
     return $self;
@@ -874,7 +1045,7 @@ that it will not cause the data already output to be changed.
 sub ship_out {
     my ($self, @objs) = @_;
 
-    return unless defined $self->{' OUTFILE'};
+    die "No output file specified" unless defined $self->{' OUTFILE'};
     my $fh = $self->{' OUTFILE'};
     seek($fh, 0, 2); # go to the end of the file
 
@@ -897,7 +1068,7 @@ sub ship_out {
         $self->{' locs'}{$objind->uid()} = $fh->tell();
         my ($objnum, $objgen) = @{$self->{' objects'}{$objind->uid()}}[0..1];
         $fh->printf('%d %d obj ', $objnum, $objgen);
-        $objind->outobjdeep($fh, $self, 'objnum' => $objnum, 'objgen' => $objgen);
+        $objind->outobjdeep($fh, $self);
         $fh->print(" endobj\n");
 
         # Note that we've output this obj, not forgetting to update
@@ -942,6 +1113,7 @@ sub copy {
                 $obj->{' objgen'} = $ng;
                 $self->add_obj($obj, $i, $ng);
                 $obj->{' parent'} = $self;
+                weaken $obj->{' parent'};
                 $obj->{' realised'} = 0;
             }
             $obj->realise;
@@ -985,7 +1157,7 @@ sub locate_obj {
 
             if ($ref->[1] == $gen) {
                 return $ref->[0] if $ref->[2] eq 'n';
-                return         # if $ref->[2] eq 'f';
+                return;        # if $ref->[2] eq 'f';
             }
         }
         $tdict = $tdict->{' prev'};
@@ -1055,6 +1227,7 @@ sub add_obj {
 
     $self->{' objcache'}{$num, $gen} = $obj;
     $self->{' objects'}{$obj->uid()} = [$num, $gen];
+    # weaken $self->{' objcache'}{$num, $gen};
     return $obj;
 }
 
@@ -1082,8 +1255,9 @@ sub _unpack_xref_stream {
     return unpack('n', $data)       if $width == 2;
     return unpack('N', "\x00$data") if $width == 3;
     return unpack('N', $data)       if $width == 4;
+    return unpack('Q>', $data)      if $width == 8;
 
-    die "Invalid column width: $width";
+    die "Unsupported xref stream entry width: $width";
 }
 
 sub readxrtr {
@@ -1133,15 +1307,13 @@ sub readxrtr {
 
         ($tdict, $buf) = $self->readval($buf);
     }
-    elsif ($buf =~ m/^(\d+)\s+(\d+)\s+obj/i)
-    {
+    elsif ($buf =~ m/^(\d+)\s+(\d+)\s+obj/i) {
         my ($xref_obj, $xref_gen) = ($1, $2);
 
         # XRef streams.
         ($tdict, $buf) = $self->readval($buf);
 
-        unless ($tdict->{' stream'})
-        {
+        unless ($tdict->{' stream'}) {
             die "Malformed XRefStm at $xref_obj $xref_gen obj in PDF file $self->{' fname'}";
         }
         $tdict->read_stream(1);
@@ -1153,12 +1325,10 @@ sub readxrtr {
         my $last;
 
         my @index;
-        if (defined $tdict->{Index})
-        {
+        if (defined $tdict->{Index}) {
             @index = map { $_->val() } @{$tdict->{Index}->val};
         }
-        else
-        {
+        else {
             @index = (0, $tdict->{Size}->val);
         }
 
@@ -1166,16 +1336,14 @@ sub readxrtr {
             $start = shift(@index);
             $last = $start + shift(@index) - 1;
 
-            for my $i ($start...$last)
-            {
+            for my $i ($start...$last) {
                 # Replaced "for $xmin" because it creates a loop-specific local variable, and we
                 # need $xmin to be correct for maxobj below.
                 $xmin = $i;
 
                 my @cols;
 
-                for my $w (@widths)
-                {
+                for my $w (@widths) {
                     my $data;
                     $data = $self->_unpack_xref_stream($w, substr($stream, 0, $w, '')) if $w;
 
@@ -1196,14 +1364,13 @@ sub readxrtr {
             }
         }
     }
-    else
-    {
+    else {
         die "Malformed xref in PDF file $self->{' fname'}";
     }
 
     $tdict->{' loc'} = $xpos;
     $tdict->{' xref'} = $xlist;
-    $self->{' maxobj'} = $xmin if $xmin > $self->{' maxobj'};
+    $self->{' maxobj'} = $xmin + 1 if $xmin + 1 > $self->{' maxobj'};
     $tdict->{' prev'} = $self->readxrtr($tdict->{'Prev'}->val)
         if (defined $tdict->{'Prev'} and $tdict->{'Prev'}->val != 0);
     delete $tdict->{' prev'} unless defined $tdict->{' prev'};
@@ -1227,19 +1394,21 @@ sub out_trailer {
         $self->ship_out();
     }
 
-    #    $size = @{$self->{' printed'}} + @{$self->{' free'}};
-    #    $tdict->{'Size'} = PDFNum($tdict->{'Size'}->val + $size);
-    # PDFSpec 1.3 says for /Size: (Required) Total number of entries in the file's
-    # cross-reference table, including the original table and all updates. Which
-    # is what the previous two lines implement.
-    # But this seems to make Acrobat croak on saving so we try the following from
-    # basil.duval@epfl.ch
+    # When writing new trailers, most dictionary entries get copied from the
+    # previous trailer, but entries related to cross-reference streams should
+    # get removed (and possibly recreated below).
+    delete $tdict->{$_} for (# Entries common to streams
+                             qw(Length Filter DecodeParms F FFilter FDecodeParms DL),
+
+                             # Entries specific to cross-reference streams
+                             qw(Index W XRefStm));
+
     $tdict->{'Size'} = PDFNum($self->{' maxobj'});
 
     my $tloc = $fh->tell();
-    $fh->print("xref\n");
+    my @out;
 
-    my @xreflist = sort { $self->{' objects'}{$a->uid}[0] <=> $self->{' objects'}{$b->uid}[0] } (@{$self->{' printed'}}, @{$self->{' free'}});
+    my @xreflist = sort { $self->{' objects'}{$a->uid}[0] <=> $self->{' objects'}{$b->uid}[0] } (@{$self->{' printed'} || []}, @{$self->{' free'} || []});
 
     my ($i, $j, $k);
     unless ($update) {
@@ -1260,34 +1429,30 @@ sub out_trailer {
         }
     }
 
-    my @freelist = sort { $self->{' objects'}{$a->uid}[0] <=> $self->{' objects'}{$b->uid}[0] } @{$self->{' free'}};
+    my @freelist = sort { $self->{' objects'}{$a->uid}[0] <=> $self->{' objects'}{$b->uid}[0] } @{$self->{' free'} || []};
 
     $j = 0; my $first = -1; $k = 0;
     for ($i = 0; $i <= $#xreflist + 1; $i++) {
-#        if ($i == 0) {
-#            $first = $i; $j = $xreflist[0]->{' objnum'};
-#            $fh->printf("0 1\n%010d 65535 f \n", $ff);
-#        }
         if ($i > $#xreflist || $self->{' objects'}{$xreflist[$i]->uid}[0] != $j + 1) {
-            $fh->print(($first == -1 ? "0 " : "$self->{' objects'}{$xreflist[$first]->uid}[0] ") . ($i - $first) . "\n");
+            push @out, ($first == -1 ? "0 " : "$self->{' objects'}{$xreflist[$first]->uid}[0] ") . ($i - $first) . "\n";
             if ($first == -1) {
-                $fh->printf("%010d 65535 f \n", defined $freelist[$k] ? $self->{' objects'}{$freelist[$k]->uid}[0] : 0);
+                push @out, sprintf("%010d 65535 f \n", defined $freelist[$k] ? $self->{' objects'}{$freelist[$k]->uid}[0] : 0);
                 $first = 0;
             }
             for ($j = $first; $j < $i; $j++) {
                 my $xref = $xreflist[$j];
-                if (defined $freelist[$k] && defined $xref && "$freelist[$k]" eq "$xref") {
+                if (defined($freelist[$k]) and defined($xref) and "$freelist[$k]" eq "$xref") {
                     $k++;
-                    $fh->print(pack("A10AA5A4",
+                    push @out, pack("A10AA5A4",
                                     sprintf("%010d", (defined $freelist[$k] ?
                                                       $self->{' objects'}{$freelist[$k]->uid}[0] : 0)), " ",
                                     sprintf("%05d", $self->{' objects'}{$xref->uid}[1] + 1),
-                                    " f \n"));
+                                    " f \n");
                 }
                 else {
-                    $fh->print(pack("A10AA5A4", sprintf("%010d", $self->{' locs'}{$xref->uid}), " ",
+                    push @out, pack("A10AA5A4", sprintf("%010d", $self->{' locs'}{$xref->uid}), " ",
                             sprintf("%05d", $self->{' objects'}{$xref->uid}[1]),
-                            " n \n"));
+                            " n \n");
                 }
             }
             $first = $i;
@@ -1297,9 +1462,53 @@ sub out_trailer {
             $j++;
         }
     }
-    $fh->print("trailer\n");
-    $tdict->outobjdeep($fh, $self);
-    $fh->print("\nstartxref\n$tloc\n%%EOF\n");
+    if (exists $tdict->{'Type'} and $tdict->{'Type'}->val() eq 'XRef') {
+        my (@index, @stream);
+        for (@out) {
+            my @a = split;
+            @a == 2 ? push @index, @a : push @stream, \@a;
+        }
+        my $i = $self->{' maxobj'}++;
+        $self->add_obj($tdict, $i, 0);
+        $self->out_obj($tdict );
+
+        push @index, $i, 1;
+        push @stream, [$tloc, 0, 'n'];
+
+        my $len = $tloc > 0xFFFF ? 4 : 2;           # don't expect files > 4 Gb
+        my $tpl = $tloc > 0xFFFF ? 'CNC' : 'CnC';   # don't expect gennum > 255, it's absurd.
+                                                    # Adobe doesn't use them anymore anyway
+        my $stream = '';
+        my @prev = (0) x ($len + 2);
+        for (@stream) {
+            $_->[1] = 0 if $_->[2] eq 'f' and $_->[1] == 65535;
+            my @line = unpack 'C*', pack $tpl, $_->[2] eq 'n' ? 1 : 0, @{$_}[0..1];
+
+            $stream .= pack 'C*', 2,                # prepend filtering method, "PNG Up"
+                map {($line[$_] - $prev[$_] + 256) % 256 } 0 .. $#line;
+            @prev = @line;
+        }
+        $tdict->{'Size'}   = PDFNum($i + 1);
+        $tdict->{'Index'}  = PDFArray(map PDFNum( $_ ), @index);
+        $tdict->{'W'}      = PDFArray(map PDFNum( $_ ), 1, $len, 1);
+        $tdict->{'Filter'} = PDFName('FlateDecode');
+
+        $tdict->{'DecodeParms'} = PDFDict();
+        $tdict->{'DecodeParms'}->val->{'Predictor'} = PDFNum(12);
+        $tdict->{'DecodeParms'}->val->{'Columns'}   = PDFNum($len + 2);
+
+        $stream = PDF::API2::Basic::PDF::Filter::FlateDecode->new->outfilt($stream, 1);
+        $tdict->{' stream'} = $stream;
+        $tdict->{' nofilt'} = 1;
+        delete $tdict->{'Length'};
+        $self->ship_out();
+    }
+    else {
+        $fh->print("xref\n", @out, "trailer\n");
+        $tdict->outobjdeep($fh, $self);
+        $fh->print("\n");
+    }
+    $fh->print("startxref\n$tloc\n%%EOF\n");
 }
 
 
