@@ -1,6 +1,9 @@
 package PDF::API2::Lite;
 
-our $VERSION = '2.025'; # VERSION
+use strict;
+no warnings qw[ deprecated recursion uninitialized ];
+
+our $VERSION = '2.044'; # VERSION
 
 BEGIN {
 
@@ -15,11 +18,9 @@ BEGIN {
 
 }
 
-no warnings qw[ deprecated recursion uninitialized ];
+=head1 NAME
 
-=head1 NAME 
-
-PDF::API2::Lite - lite pdf creation
+PDF::API2::Lite - (do not use)
 
 =head1 SYNOPSIS
 
@@ -29,6 +30,16 @@ PDF::API2::Lite - lite pdf creation
     $font = $pdf->corefont('Times-Roman');
     $font = $pdf->ttfont('TimesNewRoman.ttf');
 
+=head1 DESCRIPTION
+
+This class is unmaintained (since 2007) and should not be used in new code.  It
+combines many of the methods from L<PDF::API2> and L<PDF::API2::Content> into a
+single class but isn't otherwise any easier to use.
+
+There have been many improvements and clarifications made to the rest of the
+distribution that aren't reflected here, so the term "Lite" no longer applies.
+It remains solely for compatibility with existing legacy code.
+
 =head1 METHODS
 
 =over
@@ -94,12 +105,11 @@ deallocates the pdf-structures.
 sub saveas {
     my ($self,$file)=@_;
     if($file eq '-') {
-        return $self->{api}->stringify;
+        return $self->{api}->to_string();
     } else {
-        $self->{api}->saveas($file);
+        $self->{api}->save($file);
         return $self;
     }
-    $self->{api}->end;
     foreach my $k (keys %{$self}) {
         if(blessed($k) and $k->can('release')) {
             $k->release(1);
@@ -109,7 +119,7 @@ sub saveas {
         $self->{$k}=undef;
         delete($self->{$k});
     }
-    return undef;
+    return;
 }
 
 
@@ -558,19 +568,22 @@ sub textfont {
     return($self);
 }
 
-=item $txt->textlead $leading
+=item $txt->textleading $leading
 
 =cut
 
-sub textlead {
+# Deprecated: leading is the correct term for this operator
+sub textlead { return textleading(@_) }
+
+sub textleading {
     my $self=shift @_;
-    $self->{gfx}->lead(@_);
+    $self->{gfx}->leading(@_);
     return($self);
 }
 
 =item $pdf->text $string
 
-Applys the given text.
+Applies the given text.
 
 =cut
 
@@ -632,10 +645,4 @@ __END__
 
 =back
 
-=head1 AUTHOR
-
-alfred reibenschuh
-
 =cut
-
-
